import requests


def get_harvard_citation(doi):
    url = "https://api.crossref.org/works/" + doi + "/text"
    headers = {
        "type": "text",
        "style": "harvard3",
        "locale": "en-US",
    }
    response = requests.get(url, headers=headers, allow_redirects=True)
    response.raise_for_status()

    return response.text


def get_citation(doi):
    doi_url = f"https://doi.org/{doi}"
    print(doi_url)
    response = requests.get(
        doi_url,
        headers={
            "Accept": "application/vnd.citationstyles.csl+json",
            "style": "harvard-cite-them-right",
        },
    )
    response.raise_for_status()
    return response.text


if __name__ == "__main__":
    doi = "10.1038/nprot.2017.066"
    print(get_harvard_citation(doi))
    print(get_citation(doi))



import json
import pathlib
import shutil
from enum import Enum
from typing import Optional

import typer
from Bio import SeqIO
from typing_extensions import Annotated

from primal_page.__init__ import __version__
from primal_page.bedfiles import (
    BEDFileResult,
    BedfileVersion,
    determine_bedfile_version,
    regenerate_v3_bedfile,
    validate_bedfile,
)
from primal_page.build_index import create_index
from primal_page.dev import app as dev_app
from primal_page.download import app as download_app
from primal_page.errors import FileNotFound, InvalidReference, SchemeExists
from primal_page.modify import app as modify_app
from primal_page.modify import hashfile, regenerate_readme, trim_file_whitespace
from primal_page.schemas import (
    Collection,
    Info,
    IUPACAmbiguousDNA,
    Links,
    PrimerClass,
    SchemeStatus,
)


class FindResult(Enum):
    NOT_FOUND = 1
    FOUND = 2


# Create the typer app
app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)
app.add_typer(
    modify_app,
    name="modify",
    help="Modify an existing scheme's metadata (info.json)",
)
app.add_typer(
    download_app,
    name="download",
    help="Download schemes from the index.json",
)
app.add_typer(dev_app, name="dev", help="Development commands", hidden=True)


def typer_callback_version(value: bool):
    if value:
        typer.echo(f"primal-page version: {__version__}")
        raise typer.Exit()


@app.callback()
def primal_page(
    value: Annotated[bool, typer.Option] = typer.Option(
        False, "--version", callback=typer_callback_version
    ),
):
    pass


def validate_ref_file(ref_file: pathlib.Path):
    """
    Validate the reference.fasta file
    :param ref_file: The path to the reference.fasta file
    :raises InvalidReference: If the reference.fasta file is invalid
    """
    # Very simple fasta validation
    try:
        records = SeqIO.index(ref_file, "fasta")
    except Exception as e:
        raise InvalidReference(f"Could not validate {ref_file}: {e}") from e

    for record in records.values():
        seq_bases = set(record.seq)
        # Check DNA sequence
        if not seq_bases.issubset(IUPACAmbiguousDNA):
            raise InvalidReference(
                f"Invalid DNA bases ({', '.join(seq_bases.difference(IUPACAmbiguousDNA))}) found in {ref_file}: {record.id}"
            )
        if len(record.seq) == 0:
            raise InvalidReference(f"Empty sequence found in {ref_file}")


def find_ref(
    cli_reference: pathlib.Path | None,
    found_files: list[pathlib.Path],
    schemepath: pathlib.Path,
) -> pathlib.Path:
    """
    Find the reference.fasta file
    :param cli_reference: The reference.fasta file specified by the user. None if not specified
    :param found_files: A list of all files found in the scheme directory
    :param schemepath: The path to the scheme directory
    :return: The path to the reference.fasta file
    :raises FileNotFound: If the reference.fasta file cannot be found
    """
    # Search for reference.fasta
    if cli_reference is None:  # No reference specified
        # Search for a single *.fasta
        reference_list: list[pathlib.Path] = [
            path
            for path in found_files
            if path.name == ("reference.fasta") or path.name == ("referance.fasta")
        ]
        if len(reference_list) == 1:
            return reference_list[0]
        else:
            raise FileNotFound(
                f"Could not find a SINGLE reference.fasta file in {schemepath} or its subdirectories, found {len(reference_list)}. Please specify manually with --reference"
            )
    elif cli_reference.exists():
        validate_ref_file(cli_reference)
        return cli_reference
    else:
        raise FileNotFound(f"Could not find file at {cli_reference}")


def find_primerbed(
    cli_primerbed: pathlib.Path | None,
    found_files: list[pathlib.Path],
    schemepath: pathlib.Path,
) -> pathlib.Path:
    """
    Find the primer.bed file
    :param cli_primerbed: The primer.bed file specified by the user. None if not specified
    :param found_files: A list of all files found in the scheme directory
    :param schemepath: The path to the scheme directory
    :return: The path to the primer.bed file
    :raises FileNotFound: If the primer.bed file cannot be found
    """
    # Search for primer.bed
    if cli_primerbed is None:  # No primer.bed specified
        # Search for a single *.primer.bed
        primer_bed_list: list[pathlib.Path] = [
            path for path in found_files if path.name.endswith("primer.bed")
        ]
        if len(primer_bed_list) == 1:
            return primer_bed_list[0]
        else:
            raise FileNotFound(
                f"Could not find a SINGLE *.primer.bed file in {schemepath} or its subdirectories, found {len(primer_bed_list)}. Please specify manually with --primerbed"
            )
    elif cli_primerbed.exists():
        return cli_primerbed
    else:
        raise FileNotFound(f"Could not find file at {cli_primerbed}")


def find_config(
    cli_config: pathlib.Path | None,
    found_files: list[pathlib.Path],
    schemepath: pathlib.Path,
) -> tuple[FindResult, pathlib.Path | None]:
    """
    Find the config.json file
    :param cli_config: The config.json file specified by the user. None if not specified
    :param found_files: A list of all files found in the scheme directory
    :param schemepath: The path to the scheme directory
    :return: The path to the config.json file
    """
    # Search for config.json
    if cli_config is None:  # No config.json specified
        # Search for a single config.json
        config_list: list[pathlib.Path] = [
            path for path in found_files if path.name == ("config.json")
        ]
        match len(config_list):
            case 1:
                return (FindResult.FOUND, config_list[0])
            case _:
                return (FindResult.NOT_FOUND, None)

    elif cli_config.exists():
        # TODO validate the config.json file
        return (FindResult.FOUND, cli_config)
    else:
        return (FindResult.NOT_FOUND, None)


@app.command()
def create(
    schemepath: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to the primerscheme directory", readable=True),
    ],
    schemename: Annotated[
        str,
        typer.Option(help="The name of the scheme"),
    ],
    ampliconsize: Annotated[
        int,
        typer.Option(help="Amplicon size", min=100),
    ],
    schemeversion: Annotated[
        str,
        typer.Option(
            help="Scheme version, default is parsed from config.json",
        ),
    ],
    species: Annotated[
        list[int],
        typer.Option(
            help="The species this scheme targets. Please use NCBI taxonomy ids"
        ),
    ],
    authors: Annotated[list[str], typer.Option(help="Any authors")],
    schemestatus: Annotated[
        SchemeStatus, typer.Option(help="Scheme status")
    ] = SchemeStatus.DRAFT.value,  # type: ignore
    citations: Annotated[
        list[str], typer.Option(help="Any associated citations. Please use DOI")
    ] = [],
    primerbed: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Manually specify the primer bed file, default is *primer.bed",
            readable=True,
        ),
    ] = None,
    reference: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Manually specify the referance.fasta file, default is *.fasta",
            readable=True,
        ),
    ] = None,
    output: Annotated[
        pathlib.Path,
        typer.Option(help="Where to output the scheme", writable=True),
    ] = pathlib.Path("primerschemes"),
    configpath: Annotated[
        Optional[pathlib.Path],
        typer.Option(
            help="Where the config.json file is located", readable=True, exists=True
        ),
    ] = None,
    algorithmversion: Annotated[
        Optional[str], typer.Option(help="The version of primalscheme or other")
    ] = None,
    description: Annotated[
        Optional[str], typer.Option(help="A description of the scheme")
    ] = None,
    derivedfrom: Annotated[
        Optional[str],
        typer.Option(help="Which scheme has this scheme been derived from"),
    ] = None,
    primerclass: Annotated[
        PrimerClass, typer.Option(help="The primer class")
    ] = PrimerClass.PRIMERSCHEMES.value,  # type: ignore
    collection: Annotated[
        Optional[list[Collection]], typer.Option(help="The collection")
    ] = None,
    link_protocal: Annotated[
        list[str], typer.Option(help="Optional link to protocol")
    ] = [],
    link_validation: Annotated[
        list[str], typer.Option(help="Optional link to validation data")
    ] = [],
    links_homepage: Annotated[
        list[str], typer.Option(help="Optional link to homepage")
    ] = [],
    link_vendor: Annotated[
        list[str], typer.Option(help="Optional link to vendors")
    ] = [],
    link_misc: Annotated[
        list[str], typer.Option(help="Optional miscellaneous link")
    ] = [],
    fix: Annotated[bool, typer.Option(help="Attempt to fix the scheme")] = False,
):
    """Create a new scheme in the required format"""

    # Search for scheme repo for files
    found_files = [x for x in schemepath.rglob("*")]

    # Check for a single primer.bed file
    valid_primer_bed = find_primerbed(
        cli_primerbed=primerbed, found_files=found_files, schemepath=schemepath
    )

    match validate_bedfile(valid_primer_bed):
        case BEDFileResult.VALID:
            pass
        case BEDFileResult.INVALID_VERSION:
            raise typer.BadParameter(
                f"Could not determine primerbed version for {valid_primer_bed}"
            )
        case BEDFileResult.INVALID_STRUCTURE:
            raise typer.BadParameter(
                f"Invalid primerbed structure for {valid_primer_bed}. Please ensure it is a valid 7 column bedfile"
            )
    # Get primerbed version
    primerbed_version: BedfileVersion = determine_bedfile_version(valid_primer_bed)

    if primerbed_version != BedfileVersion.V3:
        if fix:
            try:
                bedfile_str = regenerate_v3_bedfile(valid_primer_bed)
            except Exception as e:
                raise typer.BadParameter(
                    f"Could not fix the primerbed file: {e}"
                ) from e
        else:
            raise typer.BadParameter(
                f"Primerbed version {primerbed_version.value} is not supported. Please update to v3.0 bedfile (See FAQ), or try with --fix to attempt to parse."
            )

    # Find the reference.fasta file
    valid_ref = find_ref(reference, found_files, schemepath)

    # Search for config.json
    status, conf_path = find_config(configpath, found_files, schemepath)
    config_json: None | dict = None  # type: ignore
    if status == FindResult.FOUND and conf_path is not None:  # Second check is for mypy
        configpath = conf_path
        # Read in the config
        config_json: dict = json.load(configpath.open())
        # Remove some paths from the config
        config_json.pop("output_dir", None)
        # These hashes are no longer used, so remove to prevent confusion
        md5s_keys_to_remove = {k for k in config_json.keys() if k.endswith("md5")}
        for k in md5s_keys_to_remove:
            config_json.pop(k)

        if algorithmversion is None:
            if "algorithmversion" in config_json.keys():
                algorithmversion = str(config_json["algorithmversion"])
            else:
                raise typer.BadParameter(
                    f"algorithmversion not specified in {configpath}. Please specify manually with --algorithmversion"
                )

    elif status == FindResult.NOT_FOUND:
        if algorithmversion is None:
            raise FileNotFound(
                f"Could not find a config.json file in {schemepath}. Please specify manually with --configpath or specify algorithmversion with --algorithmversion"
            )

    # Multiple pngs/htmls/msas are allowed
    # The single check is mainly to prevent multiple schemes via providing the wrong directory
    # At this point we know we have a single scheme, due to having a single primer.bed, reference.fasta, and config.json

    # Search for pngs
    pngs = [path for path in found_files if path.name.endswith(".png")]
    # Search for html
    htmls = [path for path in found_files if path.name.endswith(".html")]
    # Search for msas
    msas = [
        path
        for path in found_files
        if path.name.endswith(".fasta") and path.name != "reference.fasta"
    ]

    # Copy all additional files to working directory
    # This is done to prevent preserve the original files
    misc_files_to_copy = [
        x
        for x in found_files
        if not x.name.endswith("primer.bed")
        and not x.name.endswith("config.json")
        and not x.name.endswith("info.json")
        and not x.name.endswith(".fasta")
        and not x.name.endswith(".png")
        and not x.name.endswith(".html")
        and not x.name.endswith(".db")  # Dont copy the mismatches db
        and x.name != ".DS_Store"  # Dont copy the macos file
        and x.is_file()
    ]

    # Create the collections set
    collections = {x for x in collection} if collection is not None else set()

    # Create the links set
    links = Links(
        protocals=link_protocal,
        validation=link_validation,
        homepage=links_homepage,
        vendors=link_vendor,
        misc=link_misc,
    )
    # Create the info.json
    # Generate the md5s
    info = Info(
        ampliconsize=ampliconsize,
        schemeversion=schemeversion,
        schemename=schemename,
        primer_bed_md5="NONE",  # Will be updated later
        reference_fasta_md5="NONE",  # Will be updated later
        status=schemestatus,
        citations=set(citations),
        authors=authors,
        algorithmversion=algorithmversion,  # type: ignore
        species=set(species),
        description=description,
        derivedfrom=derivedfrom,
        primerclass=primerclass,
        articbedversion=primerbed_version,
        collections=collections,
        links=links,
    )

    #####################################
    # Final validation and create files #
    #####################################
    # Everything needs to be validated before creating files

    # Check if the repo already exists
    repo_dir = output / schemename / str(ampliconsize) / schemeversion
    if repo_dir.exists():
        raise SchemeExists(f"{repo_dir} already exists")
    repo_dir.mkdir(parents=True)

    # If this fails it will deleted the half completed scheme
    # Need to check the repo doesnt already exist
    try:
        # Copy files and trim whitespace
        if fix:
            with open(repo_dir / "primer.bed", "w") as bedfile:
                bedfile.write(bedfile_str)
        else:
            trim_file_whitespace(valid_primer_bed, repo_dir / "primer.bed")
        # parse the reference.fasta file
        with open(repo_dir / "reference.fasta", "w") as ref_file:
            records = SeqIO.parse(valid_ref, "fasta")
            SeqIO.write(records, ref_file, "fasta")

        # Update the hashes in the info.json
        info.primer_bed_md5 = hashfile(repo_dir / "primer.bed")
        info.reference_fasta_md5 = hashfile(repo_dir / "reference.fasta")

        working_dir = repo_dir / "work"
        working_dir.mkdir()

        if config_json is not None:
            # Write out the config.json
            with open(working_dir / "config.json", "w") as configfile:
                json.dump(config_json, configfile, indent=4, sort_keys=True)

        # Copy over misc files
        for misc_file in misc_files_to_copy:
            shutil.copy(misc_file, working_dir / misc_file.name)

        # Copy the pngs
        for png in pngs:
            shutil.copy(png, working_dir / png.name)
        # Copy the htmls
        for html in htmls:
            shutil.copy(html, working_dir / html.name)
        # Copy the msas
        for msa in msas:
            shutil.copy(msa, working_dir / msa.name)

        # Write info.json
        with open(repo_dir / "info.json", "w") as infofile:
            infofile.write(info.model_dump_json(indent=4))

        # Create a README.md with link to all pngs
        regenerate_readme(repo_dir, info, pngs)
    except Exception as e:
        # Cleanup
        shutil.rmtree(repo_dir)
        raise typer.BadParameter(f"{e}\nCleaning up {repo_dir}") from e


@app.command()
def build_index(
    gitaccount: Annotated[
        str,
        typer.Option(help="The name of the github account"),
    ] = "quick-lab",
    gitserver: Annotated[
        str,
        typer.Option(help="The name of the github server"),
    ] = "https://github.com/",
    parentdir: Annotated[
        pathlib.Path, typer.Option(help="The parent directory")
    ] = pathlib.Path("."),
    git_commit_sha: Annotated[
        Optional[str], typer.Option(help="The git commit")
    ] = None,
    force: Annotated[
        bool, typer.Option(help="Force the creation of the index.json")
    ] = False,
):
    """Build an index.json file from all schemes in the directory"""

    create_index(gitserver, gitaccount, parentdir, git_commit_sha, force)


@app.command()
def remove(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
):
    """Remove a scheme's version from the repo, will also remove size and schemename directories if empty"""
    # Check that this is an info.json file (for safety)
    if schemeinfo.name != "info.json":
        raise typer.BadParameter(f"{schemeinfo} is not an info.json file")

    # Remove the schemeversion directory
    shutil.rmtree(schemeinfo.parent)

    # Move up the path and remove the size directory if empty
    size_dir = schemeinfo.parent.parent
    if len(list(size_dir.iterdir())) == 0:
        size_dir.rmdir()

    # Move up the path and remove the schemename directory if empty
    scheme_dir = size_dir.parent
    if len(list(scheme_dir.iterdir())) == 0:
        scheme_dir.rmdir()


if __name__ == "__main__":
    app()



import hashlib
import json
import pathlib
import sys

from primal_page.schemas import PrimerClass


def hashfile(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def create_rawlink(repo, scheme_name, length, version, file, pclass) -> str:
    return f"https://raw.githubusercontent.com/{repo}/main/{pclass}/{scheme_name}/{length}/{version}/{file}"


def parse_version(
    version_path: pathlib.Path, repo_url, scheme_name, length, version, pclass
) -> dict[str, str]:
    version_dict = dict()

    # Read in the info.json file
    with open(version_path / "info.json") as f:
        info_dict = json.load(f)

    # Grab index.json fields
    version_dict["algorithmversion"] = info_dict["algorithmversion"]
    version_dict["status"] = info_dict["status"]
    version_dict["authors"] = info_dict["authors"]
    version_dict["citations"] = info_dict["citations"]
    version_dict["species"] = info_dict["species"]
    version_dict["license"] = info_dict["license"]
    version_dict["primerclass"] = info_dict["primerclass"]
    version_dict["schemename"] = info_dict["schemename"]
    version_dict["schemeversion"] = info_dict["schemeversion"]
    version_dict["ampliconsize"] = info_dict["ampliconsize"]
    version_dict["articbedversion"] = info_dict["articbedversion"]

    # Add a check for collections in the info.json file
    version_dict["collections"] = info_dict.get("collections", {})

    # Add the primer.bed file
    primerbed = version_path / "primer.bed"
    version_dict["primer_bed_url"] = create_rawlink(
        repo_url, scheme_name, length, version.name, primerbed.name, pclass
    )
    version_dict["primer_bed_md5"] = hashfile(primerbed)

    # Add the reference.fasta file
    reference = version_path / "reference.fasta"
    version_dict["reference_fasta_url"] = create_rawlink(
        repo_url, scheme_name, length, version.name, reference.name, pclass
    )
    version_dict["reference_fasta_md5"] = hashfile(reference)

    # Add the info.json file url
    version_dict["info_json_url"] = create_rawlink(
        repo_url, scheme_name, length, version.name, "info.json", pclass
    )

    # Check the hashes in the info.json file match the generated hashes
    if version_dict["primer_bed_md5"] != info_dict["primer_bed_md5"]:
        raise ValueError(
            f"Hash mismatch for {version_dict['primer_bed_md5']}. Expected {version_dict['primer_bed_md5']} but got {info_dict['primer_bed_md5']}"
        )
    if version_dict["reference_fasta_md5"] != info_dict["reference_fasta_md5"]:
        raise ValueError(
            f"Hash mismatch for {version_dict['reference_fasta_md5']}. Expected {version_dict['reference_fasta_md5']} but got {info_dict['reference_fasta_md5']}"
        )

    return version_dict


def parse_length(length_path, repo_url, scheme_name, length, pclass) -> dict[str, str]:
    length_dict = dict()

    # Get all the versions
    for version in length_path.iterdir():
        # Only add directories
        if not version.is_dir():
            continue

        # Parse the version
        version_dict = parse_version(
            version_path=version,
            repo_url=repo_url,
            scheme_name=scheme_name,
            length=length.name,
            version=version,
            pclass=pclass,
        )

        # Add the version to the length dict
        length_dict[version.name] = version_dict

    return length_dict


def parse_scheme(scheme_path, repo_url, scheme_name, pclass) -> dict[str, str]:
    scheme_dict = dict()

    # Get all the lengths
    for length in scheme_path.iterdir():
        # Only add directories
        if not length.is_dir():
            continue

        # Parse the length
        length_dict = parse_length(
            length_path=length,
            repo_url=repo_url,
            scheme_name=scheme_name,
            length=length,
            pclass=pclass,
        )

        # Add the length to the scheme dict
        scheme_dict[length.name] = length_dict

    return scheme_dict


def traverse_json(json_dict):
    """Depth first search of the json_dict"""
    for pclass, pclass_dict in json_dict.items():
        for scheme_name, scheme_dict in pclass_dict.items():
            for length, length_dict in scheme_dict.items():
                for version in length_dict.keys():
                    yield (pclass, scheme_name, length, version)


def check_consistency(existing_json, new_json):
    """
    Checks that paths contained in both existing_json and new_json have the same hashes (files unaltered)
    """
    # Find all paths
    existing_paths: set[tuple[str, str, str, str]] = {
        x for x in traverse_json(existing_json)
    }
    # Find all new paths
    new_paths = {x for x in traverse_json(new_json)}

    # Find all the paths that are in both
    intersection = existing_paths & new_paths

    for path in intersection:
        # Check that the reference hashes are the same
        existing_ref_hash = existing_json[path[0]][path[1]][path[2]][path[3]][
            "reference_fasta_md5"
        ]
        new_ref_hash = new_json[path[0]][path[1]][path[2]][path[3]][
            "reference_fasta_md5"
        ]
        if existing_ref_hash != new_ref_hash:
            raise ValueError(
                f"Hash changed for {path[0]}/{path[1]}/{path[2]}/{path[3]}/reference.fasta. Expected {existing_ref_hash} but got {new_ref_hash}"
            )

        # Check that the primer.bed hashes are the same
        existing_bed_hash = existing_json[path[0]][path[1]][path[2]][path[3]][
            "primer_bed_md5"
        ]
        new_bed_hash = new_json[path[0]][path[1]][path[2]][path[3]]["primer_bed_md5"]
        if existing_ref_hash != new_ref_hash:
            raise ValueError(
                f"Hash changed for {path[0]}/{path[1]}/{path[2]}/{path[3]}/primer.bed. Expected {existing_bed_hash} but got {new_bed_hash}"
            )


def create_index(
    server_url,
    repo_url,
    parent_dir=pathlib.Path("."),
    git_commit: str | None = None,
    force: bool = False,
):
    """
    Create an index JSON file for the given server and repository URLs.

    Args:
        server_url (str): The URL of the server.
        repo_url (str): The URL of the repository.
        parent_dir (str, optional): The parent directory path containing the primerscheme dir. index.json will be writem to parent_dir/index.json Defaults to ".".
        git_commit (str, optional): The git commit hash. Defaults to None.
        force (bool, optional): Force the creation of the index.json file. Allowing the change of hashes

    Returns:
        bool: True if the index JSON file is created successfully, False otherwise.
    """
    # For any Scheme, we can generate a JSON file with the following format:
    json_dict = dict()
    # Ensure the parent_dir is a pathlib.Path
    if isinstance(parent_dir, str):
        parent_dir = pathlib.Path(parent_dir)
    # Parse panels and schemes
    pclasses = [i.value for i in PrimerClass]
    for pclass in pclasses:
        # Create a dict to hold all the pclass data
        pclass_dict = dict()
        for path in (parent_dir / pclass).iterdir():
            # Only add directories
            if not path.is_dir() or path.name.startswith("."):
                continue

            # Get the Scheme name
            scheme_name = path.name
            pclass_dict[scheme_name] = parse_scheme(path, repo_url, scheme_name, pclass)

            print(f"parsed {pclass}/{scheme_name}")

        # Add the pclass to the json_dict
        json_dict[pclass] = pclass_dict

    if not force:
        # Read in the existing index.json file
        with open(parent_dir / "index.json") as f:
            existing_json_dict: dict = json.load(f)
            # Remove the github commit sha
            _old_commit = existing_json_dict.pop("github-commit-sha", None)

        # Check persistence of existing key files
        check_consistency(existing_json_dict, json_dict)

    # Update the github commit
    if git_commit is not None:
        json_dict["github-commit-sha"] = (
            git_commit  # This is added so that the index.json changes when the github commit changes
        )

    with open(parent_dir / "index.json", "w") as f:
        json.dump(json_dict, f, indent=1, sort_keys=True)

    return True


if __name__ == "__main__":
    server_url = sys.argv[1]
    repo_url = sys.argv[2]

    create_index(server_url, repo_url)



import re
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, PositiveInt
from pydantic.functional_validators import AfterValidator

from primal_page.bedfiles import BedfileVersion

INFO_SCHEMA = "v2.0.0"

SCHEMENAME_PATTERN = r"^[a-z0-9][a-z0-9-]*[a-z0-9]$"
VERSION_PATTERN = r"^v\d+\.\d+\.\d+$"
DOI_PATTERN = r"^(https:\/\/|dx.)*(doi.org\/|doi:)10.\d+\/.*$"


class PrimerClass(Enum):
    PRIMERSCHEMES = "primerschemes"


class SchemeStatus(Enum):
    WITHDRAWN = "withdrawn"
    DEPRECATED = "deprecated"
    AUTOGENERATED = "autogenerated"
    DRAFT = "draft"
    TESTING = "tested"
    VALIDATED = "validated"


def validate_schemeversion(version: str) -> str:
    if not re.match(VERSION_PATTERN, version):
        raise ValueError(
            f"Invalid version: {version}. Must match be in form of v(int).(int).(int)"
        )
    return version


def validate_schemename(schemename: str) -> str:
    if not re.match(SCHEMENAME_PATTERN, schemename):
        raise ValueError(
            f"Invalid schemename: {schemename}. Must only contain a-z, 0-9, and -. Cannot start or end with -"
        )
    return schemename


def not_empty(x: list | set | str) -> list | set | str:
    if len(x) == 0:
        raise ValueError("Cannot be empty")
    return x


IUPACAmbiguousDNA = {
    "A",
    "G",
    "K",
    "Y",
    "B",
    "S",
    "N",
    "H",
    "C",
    "W",
    "D",
    "R",
    "M",
    "T",
    "V",
}


class Collection(Enum):
    # Authors
    ARTIC = "ARTIC"
    MODJADJI = "MODJADJI"
    QUICK_LAB = "QUICK-LAB"
    COMMUNITY = "COMMUNITY"
    # Uses
    WASTE_WATER = "WASTE-WATER"
    CLINAL_ISOLATES = "CLINAL-ISOLATES"
    # Coverage types
    WHOLE_GENOME = "WHOLE-GENOME"
    PANEL = "PANEL"
    MULTI_TARGET = "MULTI-TARGET"


class Links(BaseModel):
    protocals: list[str] = []
    validation: list[str] = []
    homepage: list[str] = []
    vendors: list[str] = []
    misc: list[str] = []

    def getattr(self, link_type: str):
        """
        Get the list of links
        Raises:
        -   AttributeError if the link_type is not in the model_fields
        """
        return getattr(self, link_type)

    def append_link(
        self,
        link_type: str,
        link: str,
    ):
        """
        Add a link to the correct list
        Raises:
        -   AttributeError if the link_type is not in the model_fields
        """
        # Append the link to the correct list
        getattr(self, link_type).append(link)

    def remove_link(
        self,
        link_type: str,
        link: str,
    ):
        """
        Remove a the first occourance of the link from the correct list
        Raises:
        -   ValueError if the link is not in the list
        -   AttributeError if the link_type is not in the model_fields
        """
        getattr(self, link_type).remove(link)


class Info(BaseModel):
    ampliconsize: PositiveInt
    schemeversion: Annotated[str, AfterValidator(validate_schemeversion)]
    schemename: Annotated[str, AfterValidator(validate_schemename)]
    primer_bed_md5: str
    reference_fasta_md5: str
    status: SchemeStatus
    citations: set[str]
    authors: Annotated[list[str], AfterValidator(not_empty)]
    algorithmversion: str
    species: Annotated[set[int | str], AfterValidator(not_empty)]
    license: str = "CC BY-SA 4.0"
    primerclass: PrimerClass = PrimerClass.PRIMERSCHEMES
    infoschema: str = INFO_SCHEMA
    articbedversion: BedfileVersion
    collections: set[Collection] = set()
    links: Links = Links()
    # Add the optional fields
    description: str | None = None
    derivedfrom: str | None = None
    contactinfo: str | None = None

    def add_collection(self, collection: Collection):
        """Add a collection to the collections set"""
        self.collections.add(collection)

    def remove_collection(self, collection: Collection):
        """
        Remove a collection from the collections set.
        Raises: KeyError if the collection is not in the set
        """
        self.collections.remove(collection)

    def add_citation(self, citation: str):
        """Add a citation to the citations set"""
        self.citations.add(citation)

    def remove_citation(self, citation: str):
        """
        Remove a citation from the citations set.
        Raises: KeyError if the citation is not in the set
        """
        self.citations.remove(citation)

    def add_author(self, author: str, author_index: int | None):
        """Add an author to the authors list"""
        if author not in self.authors:
            if author_index is None:
                self.authors.append(author)
            else:
                # Insert is safe, will append if index is out of range
                self.authors.insert(author_index, author)

    def remove_author(self, author: str):
        """
        Remove an author from the authors list.
        Raises: ValueError if the author is not in the list
        """
        self.authors.remove(author)

    def reorder_authors(self, new_order: list[int]):
        """
        Reorder the authors list
        Raises:
            ValueError if duplicate indexes are found
            IndexError if the index is out of range
        """
        if len(new_order) != len(set(new_order)):
            raise ValueError("Duplicate indexes found")
            # Append authors in the new order

        new_authors = []
        for new_index in new_order:
            if new_index >= len(self.authors) or new_index < 0:
                raise IndexError(f"{new_index} is out of range")
            new_authors.append(self.authors[new_index])

        # Append any authors not in the new order
        for index, author in enumerate(self.authors):
            if index not in new_order:
                new_authors.append(author)

        self.authors = new_authors

    def change_primerclass(self, new_class: PrimerClass):
        """Change the primerclass"""
        self.primerclass = new_class

    def change_status(self, new_status: SchemeStatus):
        """Change the status"""
        self.status = new_status

    def change_license(self, new_license: str):
        """Change the license"""
        self.license = new_license

    def change_derivedfrom(self, new_derivedfrom: str):
        """Change the derivedfrom"""
        self.derivedfrom = new_derivedfrom

    def change_description(self, new_description: str):
        """Change the description"""
        self.description = new_description

    def change_contactinfo(self, new_contactinfo: str | None):
        """Change the contactinfo"""
        self.contactinfo = new_contactinfo


if __name__ == "__main__":
    info = Info(
        ampliconsize=400,
        schemeversion="v0.0.0",
        schemename="test",
        primer_bed_md5="hello",
        reference_fasta_md5="world",
        status=SchemeStatus.DRAFT,
        citations=set(),
        authors=["artic"],
        algorithmversion="test",
        species=set(["sars-cov-2"]),
        articbedversion=BedfileVersion.V3,
        collections=set(),
        links=Links(),
    )

    info_json = info.model_dump_json()
    print(info_json)

    info2 = Info.model_validate_json(info_json)



import hashlib
import json
import pathlib

import typer
from Bio import SeqIO
from typing_extensions import Annotated

from primal_page.bedfiles import (
    determine_bedfile_version,
    regenerate_v3_bedfile,
)
from primal_page.modify import hashfile, regenerate_files, trim_file_whitespace
from primal_page.schemas import (
    INFO_SCHEMA,
    BedfileVersion,
    Info,
)

app = typer.Typer(no_args_is_help=True)


@app.command()
def regenerate(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to info.json", readable=True, exists=True),
    ],
):
    """
    Regenerate the info.json and README.md file for a scheme
        - Rehashes info.json's primer_bed_md5 and reference_fasta_md5
        - Regenerates the README.md file
        - Recalculate the artic-primerbed version
        - Updates the infoschema version to current

    Ensures work/config.json has no absolute paths
        - Ensures hashes in config.json are removed
    """
    # Check that this is an info.json file (for safety)
    if schemeinfo.name != "info.json":
        raise typer.BadParameter(f"{schemeinfo} is not an info.json file")

    # Get the scheme path
    scheme_path = schemeinfo.parent

    # Get the info
    info_json = json.load(schemeinfo.open())

    # Trim whitespace from primer.bed and reference.fasta
    trim_file_whitespace(scheme_path / "primer.bed", scheme_path / "primer.bed")

    # Hash the reference.fasta file
    # If the hash is different, rewrite the file
    ref_hash = hashfile(scheme_path / "reference.fasta")
    ref_str = "".join(
        x.format("fasta") for x in SeqIO.parse(scheme_path / "reference.fasta", "fasta")
    )
    if ref_hash != hashlib.md5(ref_str.encode()).hexdigest():
        with open(scheme_path / "reference.fasta", "w") as ref_file:
            ref_file.write(ref_str)

    # if articbedversion not set then set it
    articbedversion = determine_bedfile_version(scheme_path / "primer.bed")
    if articbedversion == BedfileVersion.INVALID:
        raise typer.BadParameter(
            f"Could not determine artic-primerbed version for {scheme_path / 'primer.bed'}"
        )
    info_json["articbedversion"] = articbedversion.value

    # Regenerate the files hashes
    info_json["primer_bed_md5"] = hashfile(scheme_path / "primer.bed")
    info_json["reference_fasta_md5"] = hashfile(scheme_path / "reference.fasta")

    info = Info(**info_json)
    info.infoschema = INFO_SCHEMA

    #####################################
    # Final validation and create files #
    #####################################

    regenerate_files(info, schemeinfo)


@app.command()
def migrate(
    primerschemes: Annotated[pathlib.Path, typer.Argument(help="The parent directory")],
):
    """
    THIS MODIFIES THE SCHEMES IN PLACE. USE WITH CAUTION
        Regenerates all schemes in the primerschemes directory.
        Mainly used for migrating to the new info.json schema.
    """
    # Get all the schemes
    for schemename in primerschemes.iterdir():
        if not schemename.is_dir():
            continue
        for ampliconsize in schemename.iterdir():
            if not ampliconsize.is_dir():
                continue
            for schemeversion in ampliconsize.iterdir():
                if not schemeversion.is_dir():
                    continue
                print(f"Regenerating {schemeversion}")
                info = schemeversion / "info.json"
                if info.exists():
                    # Modify the primer.bed
                    bedfile_str = regenerate_v3_bedfile(schemeversion / "primer.bed")

                    # If the bedfile is the same, dont write it
                    if bedfile_str == (schemeversion / "primer.bed").read_text():
                        print("No changes to primer.bed")
                    else:
                        with open(schemeversion / "primer.bed", "w") as f:
                            f.write(bedfile_str)

                    # Regenerate the info.json + README.md last, as has they contain the md5s
                    regenerate(info)



import pathlib
import re
from enum import Enum

from primal_page.errors import InvalidBedFileLine, PrimerNameError, PrimerVersionError

# Primername versions
V2_PRIMERNAME = r"^[a-zA-Z0-9\-]+_[0-9]+_(LEFT|RIGHT)_[0-9]+$"
V1_PRIMERNAME = r"^[a-zA-Z0-9\-]+_[0-9]+_(LEFT|RIGHT)(_ALT[0-9]*|_alt[0-9]*)*$"

# Bedfile versions
## This doesn't parse the contents just the structure
BEDFILE_LINE = r"^\S+\t\d+\t\d+\t\S+\t\d+\t(\+|\-)\t[a-zA-Z]+$"


# bedfile versions
class BedfileVersion(Enum):
    """
    V1 bedfiles use a 6 col system
    V2 bedfiles use a 7 col system and V1 primernames
    V3 bedfiles use a 7 col system and V2 primernames
    """

    V1 = "v1.0"
    V2 = "v2.0"
    V3 = "v3.0"
    INVALID = "invalid"  # Not applicable


class BEDFileResult(Enum):
    VALID = 0
    INVALID_VERSION = 1
    INVALID_STRUCTURE = 2


class PrimerNameVersion(Enum):
    V1 = "v1"
    V2 = "v2"
    INVALID = "invalid"  # Not applicable


class BedLine:
    def __init__(
        self,
        chrom: str,
        start: int,
        end: int,
        primername: str,
        pool: int,
        strand: str,
        seq: str,
    ):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.primername = primername
        self.pool = pool
        self.strand = strand
        self.seq = seq

        # Validate the primername
        if determine_primername_version(primername) == PrimerNameVersion.INVALID:
            raise PrimerNameError(f"{primername}")

        # Autogenerated fields
        self.prefix = primername.split("_")[0]
        self.amplicon_number = int(primername.split("_")[1])
        self.pn_direction = primername.split("_")[2]
        self.primernumber: int | None = None

    def __str__(self):
        return f"{self.chrom}\t{self.start}\t{self.end}\t{self.primername}\t{self.pool}\t{self.strand}\t{self.seq}"

    def parsed_primername(self, primernumber: int) -> str:
        return "_".join(
            [
                self.prefix,
                str(self.amplicon_number),
                self.pn_direction,
                str(primernumber),
            ]
        )

    def parsed_bedline(self) -> str:
        return f"{self.chrom}\t{self.start}\t{self.end}\t{self.primername if self.primernumber is None else self.parsed_primername(self.primernumber)}\t{self.pool}\t{self.strand}\t{self.seq}"


def determine_primername_version(primername: str) -> PrimerNameVersion:
    """
    Determines the version of the primer name.

    :param primername: The primer name to check.
    :type primername: str
    :return: The version of the primer name.
    :rtype: PrimerNameVersion
    :raises: None
    """
    if re.search(V2_PRIMERNAME, primername):
        return PrimerNameVersion.V2
    elif re.search(V1_PRIMERNAME, primername):
        return PrimerNameVersion.V1
    else:
        return PrimerNameVersion.INVALID


def convert_v1_primernames_to_v2(primername: str, primernumber=1) -> str:
    """
    Convert a v1 primername to a v2 primername. Cannot convert alt primers.

    :param primername: The v1 primername to convert.
    :type primername: str
    :param primernumber: The primernumber to add to the primername.
    :type primernumber: int
    :return: The v2 primername.
    :rtype: str
    :raises: ValueError
    """
    # Check if this is a v1 primername
    if determine_primername_version(primername) != PrimerNameVersion.V1:
        raise ValueError(f"{primername} is not a valid v1 primername")

    # Split the primername
    data = primername.split("_")
    # Remove the alt
    if "alt" in data[-1].lower():
        raise ValueError(f"{primername} is a v1 alt primername, cannot convert")

    # Add primernumber and return
    data.append(str(primernumber))
    return "_".join(data)


def validate_bedfile_line_structure(line: str) -> bool:
    """
    This function validates the structure of a bedfile line, but not the contents. Empty lines will error.
    :param line: The line to validate.
    :type line: str
    :return: Whether the line is valid or not.
    :rtype: bool
    """
    line = line.strip()
    if line.startswith("#"):
        return True
    return re.search(BEDFILE_LINE, line) is not None


def validate_bedfile(bedfile: pathlib.Path) -> BEDFileResult:
    """
    This function reads in a bedfile, checks the structure of the file, and validates the primername versions.

    Args:
        bedfile (pathlib.Path): The path to the bedfile to be validated.

    Returns:
        BEDFileResult: The result of the bedfile validation. It can be one of the following:
            - BEDFileResult.INVALID_STRUCTURE: If the bedfile has an invalid structure.
            - BEDFileResult.INVALID_VERSION: If the bedfile has an invalid version.
            - BEDFileResult.VALID: If the bedfile is valid.
    """
    # Read in the bedfile string
    bedfile_str = bedfile.read_text()

    # Split the bedfile into lines
    bedlines = bedfile_str.split("\n")

    # Check each line
    for line in bedlines:
        # Skip empty lines
        if not line:
            continue
        if not validate_bedfile_line_structure(line):
            raise InvalidBedFileLine(
                f"Invalid line in bedfile:\n{line}\nShould contain 7 columns separated by tabs."
            )

    # Check the bedfile names.
    match determine_bedfile_version(bedfile):
        case BedfileVersion.INVALID:
            return BEDFileResult.INVALID_VERSION
        case _:
            return BEDFileResult.VALID


def determine_bedfile_version(input: list[list] | pathlib.Path) -> BedfileVersion:
    """
    Determine the bedfile version
    :param input: Either the bedfile lines as a list or the bedfile path
    :return: The bedfile version
    :raises:
        - PrimerNameError: If there are invalid primer names
        - PrimerVersionError: If there is a mix of primernames
    """
    if isinstance(input, pathlib.Path):
        bedlines, _ = read_bed_file(input)
    else:
        bedlines = input

    # If 6 cols then v1
    if len(bedlines[0]) == 6:
        return BedfileVersion.V1

    # If 7 cols then v2 or v3
    # Check from primername
    primernames = [x[3] for x in bedlines]
    primer_name_versions_dict = {
        x: determine_primername_version(x) for x in primernames
    }
    primer_name_versions = set(primer_name_versions_dict.values())

    if primer_name_versions == {PrimerNameVersion.V1}:
        return BedfileVersion.V2
    elif primer_name_versions == {PrimerNameVersion.V2}:
        return BedfileVersion.V3

    if PrimerNameVersion.INVALID in primer_name_versions:
        # Raise an error if there are invalid primer names
        invalid_primer_names = {
            pn
            for pn, pv in primer_name_versions_dict.items()
            if pv == PrimerNameVersion.INVALID
        }
        raise PrimerNameError(f"{invalid_primer_names}")

    if primer_name_versions == {PrimerNameVersion.V1, PrimerNameVersion.V2}:
        raise PrimerVersionError("Mix of v1 and v2 primer names")

    # This should not be reached
    return BedfileVersion.INVALID


def read_bedlines(bedfilepath: pathlib.Path) -> tuple[list[BedLine], list[str]]:
    """
    Reads a bed file and returns a list of BedLine objects.

    :return: bedfile_list
    """
    bedfile_list, bedfile_header = read_bed_file(bedfilepath)
    bedlines = []
    for line in bedfile_list:
        bedlines.append(
            BedLine(
                chrom=line[0],
                start=int(line[1]),
                end=int(line[2]),
                primername=line[3],
                pool=int(line[4]),
                strand=line[5],
                seq=line[6],
            )
        )
    return bedlines, bedfile_header


def read_bed_file(bedfilepath: pathlib.Path) -> tuple[list[list[str]], list[str]]:
    """
    Reads a bed file and returns a list of lists of strings.

    :return: bedfile_list, bedfile_header
    """

    with open(bedfilepath) as bedfile:
        bedfile_list: list[list[str]] = []
        bedfile_header: list[str] = []

        for line in bedfile:
            line = line.strip()
            # Header line
            if line.startswith("#"):
                bedfile_header.append(line)
            elif line:  # If not empty
                bedfile_list.append(line.split("\t"))

    return bedfile_list, bedfile_header


def regenerate_v3_bedfile(bedfile: pathlib.Path) -> str:
    """
    THIS FUNCTION MODIFIES THE BEDFILE. ONLY USE WHEN UPDATING THE BEDFILE VERSION.

    Convert a v2 bedfile to a v3 bedfile.
    :param bedfile: The v2 bedfile
    :return: The v3 bedfile
    """

    # Read in the bedfile
    bedlines, bedheader = read_bedlines(bedfile)

    # GUARD: If the bedfile is already v3, return the bedfile
    if determine_bedfile_version(bedfile) == BedfileVersion.V3:
        return bedfile.read_text()

    # Sort the bedlines by chrom and amplicon number
    bedlines = sorted(bedlines, key=lambda x: (x.chrom, x.amplicon_number))

    # Resolve alt primers
    # generate primerpairs
    # {chromname: {ampliconnumber: ([fbedlines], [rbedlines])}}
    sorted_bedlines: dict[str, dict[int, tuple[list[BedLine], list[BedLine]]]] = {}
    for line in bedlines:
        if line.chrom not in sorted_bedlines:
            sorted_bedlines[line.chrom] = {}

        if line.amplicon_number not in sorted_bedlines[line.chrom]:
            sorted_bedlines[line.chrom][line.amplicon_number] = ([], [])

        match line.strand:
            case "+":
                sorted_bedlines[line.chrom][line.amplicon_number][0].append(line)
            case "-":
                sorted_bedlines[line.chrom][line.amplicon_number][1].append(line)
            case _:
                raise ValueError(f"Invalid strand: {line.__str__()}")

    bedfile_str_list = bedheader

    # Calculate the primernumbers for alt primers
    for amplicon_number_dict in sorted_bedlines.values():
        for direction_list in amplicon_number_dict.values():
            fprimers, rprimers = direction_list

            # Sort and renumber the primers
            fprimers = sorted(fprimers, key=lambda x: x.seq)
            for i, line in enumerate(fprimers):
                line.primernumber = i + 1

                # Add the line to the bedfile
                bedfile_str_list.append(line.parsed_bedline())

            rprimers = sorted(rprimers, key=lambda x: x.seq)
            for i, line in enumerate(rprimers):
                line.primernumber = i + 1

                # Add the line to the bedfile
                bedfile_str_list.append(line.parsed_bedline())

    return "\n".join(bedfile_str_list) + "\n"



import hashlib
import pathlib
from typing import Optional

import typer
from typing_extensions import Annotated

from primal_page.schemas import (
    Collection,
    Info,
    Links,
    PrimerClass,
    SchemeStatus,
)

LICENSE_TXT_CC_BY_SA_4_0 = """\n\n------------------------------------------------------------------------

This work is licensed under a [Creative Commons Attribution-ShareAlike 4.0 International License](http://creativecommons.org/licenses/by-sa/4.0/) 

![](https://i.creativecommons.org/l/by-sa/4.0/88x31.png)"""

app = typer.Typer(no_args_is_help=True)


def trim_file_whitespace(in_path: pathlib.Path, out_path: pathlib.Path):
    """
    Trim whitespace from the ends of a file.
        - Reads file into memory. Not suitable for large files
    """
    inlines = []
    with open(in_path) as infile:
        for line in infile:
            inlines.append(line.strip() + "\n")

    with open(out_path, "w") as outfile:
        outfile.writelines(inlines)


def hashfile(fname: pathlib.Path) -> str:
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def regenerate_readme(path: pathlib.Path, info: Info, pngs: list[pathlib.Path]):
    """
    Regenerate the README.md file for a scheme

    :param path: The path to the scheme directory
    :type path: pathlib.Path
    :param info: The scheme information
    :type info: Info
    :param pngs: The list of PNG files
    :type pngs: list[pathlib.Path]
    """

    with open(path / "README.md", "w") as readme:
        readme.write(
            f"# {info.schemename} {info.ampliconsize}bp {info.schemeversion}\n\n"
        )

        readme.write(
            f"[primalscheme labs](https://labs.primalscheme.com/detail/{info.schemename}/{info.ampliconsize}/{info.schemeversion})\n\n"
        )

        if info.description is not None:
            readme.write("## Description\n\n")
            readme.write(f"{info.description}\n\n")

        readme.write("## Overviews\n\n")
        for png in pngs:
            readme.write(f"![{png.name}](work/{png.name})\n\n")

        readme.write("## Details\n\n")

        # Write the detials into the readme
        readme.write(f"""```json\n{info.model_dump_json(indent=4)}\n```\n\n""")

        if info.license == "CC BY-SA 4.0":
            readme.write(LICENSE_TXT_CC_BY_SA_4_0)


def regenerate_files(info: Info, schemeinfo: pathlib.Path):
    # Write the validated info.json
    with open(schemeinfo, "w") as infofile:
        infofile.write(info.model_dump_json(indent=4))

    # Update the README
    scheme_path = schemeinfo.parent
    pngs = [path for path in scheme_path.rglob("*.png")]
    regenerate_readme(scheme_path, info, pngs)


@app.command()
def add_link(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    linkfield: Annotated[
        str,
        typer.Argument(
            help=f"The link field to add to. {', '.join(Links.model_fields.keys())}"
        ),
    ],
    link: Annotated[
        str,
        typer.Argument(help="The link to add."),
    ],
):
    """
    Add a link to the selected link field to the info.json
    """
    info = Info.model_validate_json(schemeinfo.read_text())

    try:
        info.links.append_link(linkfield, link)
    except AttributeError:
        raise typer.BadParameter(
            f"{linkfield} is not a valid link field. Please choose from {', '.join(Links.model_fields.keys())}"
        ) from None

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def remove_link(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    linkfield: Annotated[
        str,
        typer.Argument(
            help=f"The link field to remove from. {', '.join(Links.model_fields.keys())}"
        ),
    ],
    link: Annotated[
        str,
        typer.Argument(help="The link to remove."),
    ],
):
    """
    Add a link to the selected link field to the info.json
    """
    info = Info.model_validate_json(schemeinfo.read_text())

    try:
        info.links.remove_link(linkfield, link)
    except AttributeError:
        raise typer.BadParameter(
            f"{linkfield} is not a valid link field. Please choose from {', '.join(Links.model_fields.keys())}"
        ) from None
    except ValueError:
        raise typer.BadParameter(
            f"{link} is not in links[{linkfield}]: {info.links.getattr(linkfield)}"
        ) from None

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def add_author(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    author: Annotated[str, typer.Argument(help="The author to add")],
    author_index: Annotated[
        Optional[int],
        typer.Option(
            help="The 0-based index to insert the author at. Default is the end"
        ),
    ],
):
    """Append an author to the authors list in the info.json file"""

    info = Info.model_validate_json(schemeinfo.read_text())

    info.add_author(author, author_index)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def remove_author(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    author: Annotated[str, typer.Argument(help="The author to remove")],
):
    """Remove an author from the authors list in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    try:
        info.remove_author(author)
    except KeyError:
        raise typer.BadParameter(f"{author} is already not present") from None

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def reorder_authors(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    author_index: Annotated[
        Optional[str],
        typer.Argument(
            help="The indexes in the new order, seperated by spaces. e.g. 1 0 2. Any indexes not provided will be appended to the end"
        ),
    ] = None,
):
    """Reorder the authors in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Reorder interactively
    if author_index is None:
        # Current order
        typer.echo("Current order:")
        for index, author in enumerate(info.authors):
            typer.echo(f"{index}: {author}")

        # Get the new order
        new_order_str: str = typer.prompt(
            "Please provide the indexes in the new order, seperated by spaces. e.g. 1 0 2. Any indexes not provided will be appended to the end",
            type=str,
        )
        new_order = [int(x) for x in new_order_str.split()]
    else:  # Reorder via cli
        new_order = [int(x) for x in author_index.split()]

    try:
        info.reorder_authors(new_order)
    except ValueError as e:
        raise typer.BadParameter(f"{e}") from None
    except IndexError as e:
        raise typer.BadParameter(f"{e}") from None

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def add_citation(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    citation: Annotated[str, typer.Argument(help="The citation to add")],
):
    """Append an citation to the authors list in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Add the citation
    info.add_citation(citation)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def remove_citation(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    citation: Annotated[str, typer.Argument(help="The citation to remove")],
):
    """Remove an citation form the authors list in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_bytes())

    try:
        info.remove_citation(citation)
    except KeyError:
        raise typer.BadParameter(f"{citation} is already not present") from None

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def remove_collection(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    collection: Annotated[Collection, typer.Argument(help="The Collection to remove")],
):
    """Remove an Collection from the Collection list in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Check if collection is already not in the list
    try:
        info.remove_collection(collection)
    except KeyError:
        raise typer.BadParameter(f"{collection} is already not present") from None

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def add_collection(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    collection: Annotated[Collection, typer.Argument(help="The Collection to add")],
):
    """Add a Collection to the Collection list in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    info.add_collection(collection)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def change_description(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    description: Annotated[
        str,
        typer.Argument(
            help="The new description. Use 'None' to remove the description"
        ),
    ],
):
    """Replaces the description in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Change the description
    info.change_description(description.strip())

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def change_derivedfrom(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    derivedfrom: Annotated[
        str,
        typer.Argument(
            help="The new derivedfrom. Use 'None' to remove the derivedfrom"
        ),
    ],
):
    """Replaces the derivedfrom in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Add the derivedfrom
    info.change_derivedfrom(derivedfrom.strip())

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def change_license(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    license: Annotated[
        str,
        typer.Argument(
            help="The new license. Use 'None' show the work is not licensed (Not recommended)"
        ),
    ],
):
    """Replaces the license in the info.json file"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Change the license
    info.change_license(license)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def change_status(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    schemestatus: Annotated[
        SchemeStatus,
        typer.Argument(
            help="The scheme class",
        ),
    ] = SchemeStatus.DRAFT,
):
    """Change the status field in the info.json"""

    info = Info.model_validate_json(schemeinfo.read_text())

    # Change the status
    info.change_status(schemestatus)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def change_primerclass(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    primerclass: Annotated[
        PrimerClass, typer.Argument(help="The primerclass to change to")
    ],
):
    """Change the primerclass field in the info.json"""

    info = Info.model_validate_json(schemeinfo.read_text())

    # Change the primerclass
    info.change_primerclass(primerclass)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)


@app.command()
def change_contactinfo(
    schemeinfo: Annotated[
        pathlib.Path,
        typer.Argument(
            help="The path to info.json", readable=True, exists=True, writable=True
        ),
    ],
    contactinfo: Annotated[
        Optional[str],
        typer.Argument(
            help="The contact infomation for this scheme. Use 'None' to remove the contact info",
        ),
    ],
):
    """Change the contactinfo field in the info.json"""
    info = Info.model_validate_json(schemeinfo.read_text())

    # Change the contactinfo
    info.change_contactinfo(contactinfo)

    # Write the validated info.json and regenerate the README
    regenerate_files(info, schemeinfo)



from click import UsageError


class PrimerNameError(UsageError):
    """Raised when a primername is invalid"""

    def __init__(self, primername: str):
        super().__init__(
            f"Invalid primernames: {primername}. Please use format (name)_(amplicon-number)_(LEFT|RIGHT) with optional _(primer-number)"
        )


class PrimerVersionError(UsageError):
    """Raised when a primername is unexpected"""

    def __init__(self, message: str):
        super().__init__(message)


class InvalidBedFileLine(UsageError):
    """Raised when a bedline is invalid"""

    def __init__(self, message: str):
        super().__init__(message)


class SchemeExists(UsageError):
    """Raised when a Scheme already exists"""

    def __init__(self, message: str):
        super().__init__(message)


class FileNotFound(UsageError):
    """Raised when a file is not found"""

    def __init__(self, message: str):
        super().__init__(message)


class InvalidReference(UsageError):
    """Raised when a file is not found"""

    def __init__(self, message: str):
        super().__init__(message)



import hashlib
import json
import pathlib

import requests
import typer
from typing_extensions import Annotated

from primal_page.schemas import validate_schemename, validate_schemeversion

app = typer.Typer(no_args_is_help=True)


def validate_hashes(input_text: str, expected_hash: str, output_file: pathlib.Path):
    """
    Validate the hash of the input text. If the hash does not match the expected hash, raise a ValueError.
    If the hash does match, write the input text to the output file.
    """
    input_hash = hashlib.md5(input_text.encode()).hexdigest()
    if input_hash != expected_hash:
        raise ValueError(
            f"WARNING: HASH MISMATCH: Expected {expected_hash} but got {input_hash}. File not saved to disk."
        )

    # Write the file
    with open(output_file, "w") as f:
        f.write(input_text)


def download_scheme_func(
    schemename: str,
    ampliconsize: str,
    schemeversion: str,
    index: dict,
    output_dir: pathlib.Path,
):
    # Grab the primerschemes
    primerschemes = index.get("primerschemes", {})

    # Grab the scheme
    scheme = (
        primerschemes.get(schemename, {}).get(ampliconsize, {}).get(schemeversion, {})
    )
    if not scheme:
        raise ValueError(
            f"Scheme {schemename}/{ampliconsize}/{schemeversion} not found in index.json"
        )

    # Create the output directory
    scheme_dir = output_dir / schemename / str(ampliconsize) / schemeversion
    scheme_dir.mkdir(parents=True, exist_ok=True)

    # Download the bedfile
    bedfile_url = scheme["primer_bed_url"]
    bedfile_text = requests.get(bedfile_url).text

    # Validate the hash before write the file
    validate_hashes(bedfile_text, scheme["primer_bed_md5"], scheme_dir / "primer.bed")

    # Download the reference
    reference_url = scheme["reference_fasta_url"]
    reference_text = requests.get(reference_url).text

    # Validate the hash before write the file
    validate_hashes(
        reference_text,
        scheme["reference_fasta_md5"],
        scheme_dir / "reference.fasta",
    )

    # Download the info.json
    info_url = scheme["info_json_url"]
    info_text = requests.get(info_url).text
    # Write the file
    with open(scheme_dir / "info.json", "w") as f:
        f.write(info_text)

    print(f"Downloaded:\t{schemename}/{ampliconsize}/{schemeversion}")


def fetch_index(index_url: str) -> dict:
    """Download the index.json and return it as a dict"""
    try:
        r = requests.get(index_url)
        r.raise_for_status()
        index_json = json.loads(r.text)
        return index_json
    except requests.exceptions.HTTPError as err:
        raise err
    except json.JSONDecodeError as err:
        raise err


def download_all_func(index: dict, output: pathlib.Path):
    """Download all schemes from the index.json"""

    # Grab the primerschemes
    primerschemes = index.get("primerschemes", {})

    # Download all the schemes
    # Create a list of all schemes
    for schemename in primerschemes:
        for ampliconsize in primerschemes[schemename]:
            for schemeversion in primerschemes[schemename][ampliconsize]:
                download_scheme_func(
                    schemename, ampliconsize, schemeversion, index, output
                )


@app.command()
def download_all(
    output: Annotated[
        pathlib.Path,
        typer.Option(help="The directory the primerschemes dir will be created in"),
    ],
    index_url: Annotated[
        str, typer.Option(help="The URL to the index.json")
    ] = "https://raw.githubusercontent.com/quick-lab/primerschemes/main/index.json",
):
    """Download all schemes from the index.json"""
    # Fetch the index and store in memory
    index = fetch_index(index_url)

    # Create the output directory
    output_primerschemes = output / "primerschemes"
    output_primerschemes.mkdir(exist_ok=True)

    download_all_func(output=output, index=index)


@app.command()
def download_scheme(
    schemename: Annotated[
        str,
        typer.Argument(help="The name of the scheme", callback=validate_schemename),
    ],
    ampliconsize: Annotated[int, typer.Argument(help="Amplicon size")],
    schemeversion: Annotated[
        str, typer.Argument(help="Scheme version", callback=validate_schemeversion)
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(help="The directory the primerschemes dir will be created in"),
    ],
    index_url: Annotated[
        str, typer.Option(help="The URL to the index.json")
    ] = "https://raw.githubusercontent.com/quick-lab/primerschemes/main/index.json",
):
    """Download a scheme from the index.json"""

    # Fetch the index and store in memory
    index = fetch_index(index_url)

    # Create the output directory
    output_primerschemes = output / "primerschemes"
    output_primerschemes.mkdir(exist_ok=True)

    download_scheme_func(
        output_dir=output_primerschemes,
        index=index,
        schemename=schemename,
        ampliconsize=str(ampliconsize),
        schemeversion=schemeversion,
    )
